from .ply import yacc
from .lex import *
from .ast import *
from .exception import *


# Precedence rules
precedence = (
    ('left', 'OR', 'AND'),
    ('left', 'LESS', 'GREATER', 'LESSEQUAL', 'GREATEREQUAL', 'EQEQUAL', 'NOTEQUAL'),
    ('left', 'PLUS', 'MINUS'),
    ('left', 'STAR', 'SLASH', 'DOUBLESLASH', 'PERCENT'),
    ('left', 'DOUBLESTAR'),
    ('right', 'NOT', 'UADD', 'USUB')
)

# Rule for a program
def p_program(p):
    '''program : statements'''
    statements = StatementSequence(p[1])
    p[0] = Program(statements)

# Rule for statement sequence
def p_statements(p):
    '''statements : statements statement
                  | statement'''
    if (len(p) == 2 and p[1]):
        statement = p[1]
        p[0] = [statement]
    elif (len(p) == 3):
        statements = p[1]
        if (p[2]): statements.append(p[2])
        p[0] = statements

# Statement rule for function definition
def p_statement_functiondef(p):
    '''statement : FUNCTIONDEF NAME LPAR arguments_comma RPAR BEGIN statements END'''
    name = Name(p[2])
    arguments = ArgumentSequence(p[4])
    body = StatementSequence(p[7])
    p[0] = FunctionDef(name, arguments, body)

# Rule for argument sequence separated by comma
def p_arguments_comma(p):
    '''arguments_comma : arguments_comma COMMA NAME
                       | NAME
                       | empty'''
    if (len(p) == 2 and p[1]):
        argument = Name(p[1])
        p[0] = [argument]
    elif (len(p) == 4):
        arguments = p[1]
        if (p[3]): arguments.append(Name(p[3]))
        p[0] = arguments
    else:
        p[0] = []

# Rule for return
def p_statement_return(p):
    '''statement : RETURN expression'''
    p[0] = Return(p[2])

# Statement rule for assignment
def p_statement_assignment(p):
    '''statement : NAME EQUAL expression'''
    name = Name(p[1])
    p[0] = Assignment(name, p[3])

# Statement rule for if-then
def p_statement_ifthen(p):
    '''statement : IF expression THEN statements elifs END'''
    expression = p[2]
    statements = StatementSequence(p[4])
    elifs = ElifSequence(p[5])
    p[0] = If(expression, statements, elifs)

# Statement rule for if-then-else
def p_statement_ifthenelse(p):
    '''statement : IF expression THEN statements elifs ELSE statements END'''
    expression = p[2]
    statements_if = StatementSequence(p[4])
    statements_else = StatementSequence(p[7])
    elifs = ElifSequence(p[5], statements_else)
    p[0] = If(expression, statements_if, elifs)

# Rule for elif sequence
def p_elifs(p):
    '''elifs : elifs ELIF expression THEN statements
             | ELIF expression THEN statements
             | empty'''
    if (len(p) == 5):
        expression = p[2]
        statements = StatementSequence(p[4])
        p[0] = [(expression, statements)]
    elif (len(p) == 6):
        elifs = p[1]
        expression = p[3]
        statements = StatementSequence(p[5])
        if (p[2]): elifs.append((expression, statements))
        p[0] = elifs
    else:
        p[0] = []

# Statement rule for iterate
def p_statement_iterate(p):
    '''statement : FOR NAME ITERATE NAME BEGIN statements END'''
    variable = p[2]
    iterable = Name(p[4])
    statements = StatementSequence(p[6])
    p[0] = Iterate(variable, iterable, statements)

# Statement rule for loop
def p_statement_loop(p):
    '''statement : LOOP expression BEGIN statements END'''
    expression = p[2]
    statements = StatementSequence(p[4])
    p[0] = Loop(expression, statements)

# Statement rule for continue
def p_statement_continue(p):
    '''statement : CONTINUE'''
    p[0] = Continue()

# Statement rule for break
def p_statement_break(p):
    '''statement : BREAK'''
    p[0] = Break()

# Statement rule for expression
def p_statement_expression(p):
    '''statement : expression'''
    p[0] = p[1]

# Expression rule for group surrounded with parentheses
def p_expression_group(p):
    '''expression : LPAR expression RPAR'''
    p[0] = p[2]

# Rule for expression sequence separated by comma
def p_expressions_comma(p):
    '''expressions_comma : expressions_comma COMMA expression
                         | expression
                         | empty'''
    if (len(p) == 2 and p[1]):
        expression = p[1]
        p[0] = [expression]
    elif (len(p) == 4):
        expressions = p[1]
        if (p[3]): expressions.append(p[3])
        p[0] = expressions
    else:
        p[0] = []

# Expression rule for unary operator
def p_expression_unary_operators(p):
    '''expression : NOT expression %prec NOT
                  | PLUS expression %prec UADD
                  | MINUS expression %prec USUB'''
    value = p[2]
    operator = EXACT_TOKEN_TYPES[p[1]]
    if (operator == 'NOT'):
        p[0] = Not(value)
    elif (operator == 'PLUS'):
        p[0] = UAdd(value)
    elif (operator == 'MINUS'):
        p[0] = USubtract(value)

# Expression rule for binary operator
def p_expression_binary_operators(p):
    '''expression : expression PLUS expression
                  | expression MINUS expression
                  | expression STAR expression
                  | expression SLASH expression
                  | expression PERCENT expression
                  | expression DOUBLESTAR expression
                  | expression DOUBLESLASH expression'''
    left = p[1]
    right = p[3]
    operator = EXACT_TOKEN_TYPES[p[2]]
    if (operator == 'PLUS'):
        p[0] = Add(left, right)
    elif (operator == 'MINUS'):
        p[0] = Subtract(left, right)
    elif (operator == 'STAR'):
        p[0] = Multiply(left, right)
    elif (operator == 'SLASH'):
        p[0] = Divide(left, right)
    elif (operator == 'PERCENT'):
        p[0] = Modulo(left, right)
    elif (operator == 'DOUBLESTAR'):
        p[0] = Power(left, right)
    elif (operator == 'DOUBLESLASH'):
        p[0] = FloorDivide(left, right)

# Expression rule for boolean operator
def p_expression_boolean_operators(p):
    '''expression : expression AND expression
                  | expression OR expression'''
    left = p[1]
    right = p[3]
    operator = EXACT_TOKEN_TYPES[p[2]]
    if (operator == 'AND'):
        p[0] = And(left, right)
    elif (operator == 'OR'):
        p[0] = Or(left, right)

# Expression rule for compare operator
def p_expression_compare_operators(p):
    '''expression : expression EQEQUAL expression
                  | expression NOTEQUAL expression
                  | expression LESS expression
                  | expression LESSEQUAL expression
                  | expression GREATER expression
                  | expression GREATEREQUAL expression'''
    left = p[1]
    right = p[3]
    operator = EXACT_TOKEN_TYPES[p[2]]
    if (operator == 'EQEQUAL'):
        p[0] = Equal(left, right)
    elif (operator == 'NOTEQUAL'):
        p[0] = NotEqual(left, right)
    elif (operator == 'LESS'):
        p[0] = Less(left, right)
    elif (operator == 'LESSEQUAL'):
        p[0] = LessEqual(left, right)
    elif (operator == 'GREATER'):
        p[0] = Greater(left, right)
    elif (operator == 'GREATEREQUAL'):
        p[0] = GreaterEqual(left, right)

# Expression rule for boolean constant
def p_expression_boolean(p):
    '''expression : TRUE
                  | FALSE'''
    constant = EXACT_TOKEN_TYPES[p[1]]
    if (constant == 'TRUE'):
        p[0] = Boolean(True)
    elif (constant == 'FALSE'):
        p[0] = Boolean(False)

# Expression rule for integer number
def p_expression_integer(p):
    '''expression : INTEGER'''
    p[0] = Integer(p[1])

# Expression rule for float number
def p_expression_float(p):
    '''expression : FLOAT'''
    p[0] = Float(p[1])

# Expression rule for complex number
def p_expression_complex(p):
    '''expression : COMPLEX'''
    p[0] = Complex(p[1])

# Expression rule for string
def p_expression_string(p):
    '''expression : STRING'''
    p[0] = String(p[1])

# Expression rule for variable name
def p_expression_name(p):
    '''expression : NAME'''
    p[0] = Name(p[1])

# Expression rule for function call
def p_expression_functioncall(p):
    '''expression : NAME LPAR expressions_comma RPAR'''
    name = Name(p[1])
    expressions = ExpressionSequence(p[3])
    p[0] = FunctionCall(name, expressions)

# Expression rule for list
def p_expression_list(p):
    '''expression : LSQB expressions_comma RSQB'''
    expressions = ExpressionSequence(p[2])
    p[0] = ArrayList(expressions)

# Expression rule for iterable index
def p_expression_iterable_index(p):
    '''expression : NAME INDEX expression'''
    name = Name(p[1])
    expression = p[3]
    p[0] = IterableIndex(name, expression)

# Rule for nothing
def p_empty(p):
    '''empty : '''

# Error rule for syntax error
def p_error(p):
    raise SyntaxException(p)

# Build the parser
parser = yacc.yacc()



import sys
import inspect
from .lex import *
from .parse import *
from .ast import *
from .symtable import *
from .builtins import *


def exec(data, debug=False):
    if (debug):
        print('[DEBUG] input data as raw string:')
        print(repr(data))
        print('[DEBUG] input data as string:')
        print(data)

    # Lex
    lexer.input(data)

    if (debug):
        print('[DEBUG] tokens:')
        while (True):
            tok = lexer.token()
            if (not tok): break
            print(tok)
    
    if (debug):
        print('[DEBUG] program output:')
    
    # initialise symbol table
    symtable = SymbolTable()

    # Add builtin functions to the symbol table
    builtins = {
        name: obj for name, obj in inspect.getmembers(sys.modules['src.sheng.builtins'])
    }
    for identifier, builtin_name in BUILTIN_NAMES.items():
        name = Name(identifier)
        function_pointer = builtins[builtin_name]
        arguments = ArgumentSequence(
            [Name(arg) for arg in list(function_pointer.__code__.co_varnames)]
        )
        functiondef = BuiltinFunctionDef(name, arguments, function_pointer)
        symtable.insert(identifier, functiondef)
    
    # Parse
    ast = parser.parse(data)
    ast.evaluate(symtable)



# TODO

def builtin_print(value):
    if (isinstance(value, str)):
        # Discard leading and trailing quotes
        value = value[1:-1]
    if (isinstance(value, bool)):
        value = '对' if value else '错'
    print(value)
    return None



from abc import *
from .symtable import *
from .exception import *


class Node(ABC):

    def evaluate(self, symtable):
        raise NotImplementedError()


class NodeSequence(Node):

    def __init__(self, sequence):
        self._sequence = sequence
        self._index = -1

    def __getitem__(self, index):
        return self._sequence[index]

    def __len__(self):
        return len(self._sequence)

    def __iter__(self):
        return iter(self._sequence)


class Program(Node):

    def __init__(self, statements):
        self.statements = statements

    def evaluate(self, symtable):
        return self.statements.evaluate(symtable)


class Statement(Node): pass


class StatementSequence(NodeSequence):

    def evaluate(self, symtable):
        for node in self._sequence:
            value = node.evaluate(symtable)
            if (value):
                return value
        return None


class FunctionDef(Statement):

    def __init__(self, name, arguments, body):
        self.name = name
        self.arguments = arguments
        self.body = body

    def evaluate(self, symtable):
        identifier = self.name.get_identifier()
        symtable.insert(identifier, self)
        return None

    def get_arguments(self):
        return self.arguments

    def get_body(self):
        return self.body


class BuiltinFunctionDef(FunctionDef):

    def __init__(self, name, arguments, function_pointer):
        super().__init__(
            name, arguments, 
            BuiltinFunctionBody(function_pointer, arguments)
        )


class BuiltinFunctionBody(Node):

    def __init__(self, function_pointer, arguments):
        self.function_pointer = function_pointer
        self.arguments = arguments

    def evaluate(self, symtable):
        args = {}
        for argument in self.arguments:
            value = argument.evaluate(symtable)
            args[argument.get_identifier()] = value
        return self.function_pointer(**args)


class ArgumentSequence(NodeSequence): pass


class Return(Statement):

    def __init__(self, expression):
        self.expression = expression

    def evaluate(self, symtable):
        value = self.expression.evaluate(symtable)
        ReturnValue = type('ReturnValue', (type(value), ), {})
        return ReturnValue(value)


class Assignment(Statement):

    def __init__(self, name, node):
        self.name = name
        self.node = node

    def evaluate(self, symtable):
        identifier = self.name.get_identifier()
        node = Constant(self.node.evaluate(symtable))
        symtable.insert(identifier, node)
        return None


class If(Statement):

    def __init__(self, expression, statements, elifs):
        self.expression = expression
        self.statements = statements
        self.elifs = elifs

    def evaluate(self, symtable):
        if (self.expression.evaluate(symtable)):
            return self.statements.evaluate(symtable)
        else:
            return self.elifs.evaluate(symtable)


class ElifSequence(NodeSequence):

    def __init__(self, sequence, statements_else=None):
        super().__init__(sequence)
        self.statements_else = statements_else

    def evaluate(self, symtable):
        for expression, statements in self._sequence:
            if (expression.evaluate(symtable)):
                return statements.evaluate(symtable)
        if (self.statements_else):
            return self.statements_else.evaluate(symtable)
        return None


class Iterate(Statement):

    def __init__(self, variable, iterable, statements):
        self.variable = variable
        self.iterable = iterable
        self.statements = statements

    def evaluate(self, symtable):
        for e in self.iterable.evaluate(symtable):
            node = Constant(e)
            symtable.insert(self.variable, node)
            value = self.statements.evaluate(symtable)
            if (type(value).__name__ == 'ReturnValue'):
                return value
            if (type(value).__name__ == 'ContinueValue'):
                continue
            if (type(value).__name__ == 'BreakValue'):
                break
        return None


class Loop(Statement):

    def __init__(self, expression, statements):
        self.expression = expression
        self.statements = statements

    def evaluate(self, symtable):
        while (self.expression.evaluate(symtable)):
            value = self.statements.evaluate(symtable)
            if (type(value).__name__ == 'ReturnValue'):
                return value
            if (type(value).__name__ == 'ContinueValue'):
                continue
            if (type(value).__name__ == 'BreakValue'):
                break
        return None


class Continue(Statement):

    def evaluate(self, symtable):
        ContinueValue = type('ContinueValue', (object, ), {})
        return ContinueValue()


class Break(Statement):

    def evaluate(self, symtable):
        BreakValue = type('BreakValue', (object, ), {})
        return BreakValue()


class Expression(Node): pass


class ExpressionSequence(NodeSequence): pass


class Constant(Expression):

    def __init__(self, value):
        self.value = value

    def evaluate(self, symtable):
        return self.value


class Boolean(Constant):

    def evaluate(self, symtable):
        return bool(self.value)


class Integer(Constant):

    def evaluate(self, symtable):
        return int(self.value)


class Float(Constant):

    def evaluate(self, symtable):
        return float(self.value)


class Complex(Constant):

    def evaluate(self, symtable):
        return complex(self.value)


class String(Constant):

    def evaluate(self, symtable):
        return str(self.value)


class UnaryOp(Expression):

    def __init__(self, node):
        self.node = node


class Not(UnaryOp):

    def evaluate(self, symtable):
        value = self.node.evaluate(symtable)
        return not value


class USubtract(UnaryOp):

    def evaluate(self, symtable):
        value = self.node.evaluate(symtable)
        return +value


class UAdd(UnaryOp):

    def evaluate(self, symtable):
        value = self.node.evaluate(symtable)
        return -value


class BinaryOp(Expression):

    def __init__(self, left, right):
        self.left = left
        self.right = right


class Add(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue + rvalue


class Subtract(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue - rvalue


class Multiply(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue * rvalue


class Divide(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue / rvalue


class Modulo(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue % rvalue


class Power(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue ** rvalue


class FloorDivide(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue // rvalue


class BoolOp(BinaryOp): pass


class And(BoolOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue and rvalue


class Or(BoolOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue or rvalue


class CompareOp(BinaryOp): pass


class Equal(CompareOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue == rvalue


class NotEqual(CompareOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue != rvalue


class Less(CompareOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue < rvalue


class LessEqual(CompareOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue <= rvalue


class Greater(CompareOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue > rvalue


class GreaterEqual(CompareOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return lvalue >= rvalue


class Name(Expression):

    def __init__(self, value):
        self.value = value

    def get_identifier(self):
        return self.value
    
    def evaluate(self, symtable):
        symbol = symtable.lookup(self.value)
        node = symbol.get_node()
        return node.evaluate(symtable)


class FunctionCall(Expression):

    def __init__(self, name, expressions):
        self.name = name
        self.expressions = expressions
    
    def evaluate(self, symtable):
        identifier = self.name.get_identifier()
        symbol = symtable.lookup(identifier)
        if (not isinstance(symbol.get_node(), FunctionDef)):
            raise RuntimeException(f'{identifier} is not a function')

        # Get function definition and its attributes
        functiondef = symbol.get_node()
        arguments = functiondef.get_arguments()
        body = functiondef.get_body()

        # Create child symbol table and insert parent symbols
        function_symtable = Function()
        symbols = symtable.to_dict()
        for _, s in symbols.items():
            function_symtable.insert(s.get_identifier(), s.get_node())

        # Assign values to arguments respectively
        for i in range(len(self.expressions)):
            arg_identifier = arguments[i].get_identifier()
            arg_node = Constant(self.expressions[i].evaluate(symtable))
            function_symtable.insert(arg_identifier, arg_node)
        
        # Evaluate statements in body
        value = body.evaluate(function_symtable)
        if (type(value).__name__ == 'ReturnValue'):
            return value
        return None


class Iterable(Constant): pass


class ArrayList(Iterable):

    def __init__(self, expressions):
        self.expressions = expressions
    
    def evaluate(self, symtable):
        return [
            expression.evaluate(symtable) 
            for expression in self.expressions
        ]


class IterableIndex(Expression):

    def __init__(self, name, expression):
        self.name = name
        self.expression = expression

    def evaluate(self, symtable):
        iterable = self.name.evaluate(symtable)
        index = self.expression.evaluate(symtable)
        return iterable[index]



from typing import *
from .exception import *


class SymbolTable:

    def __init__(self):
        self.symbols = dict()
    
    def insert(self, identifier, node):
        symbol = Symbol(identifier, node)
        self.symbols[identifier] = symbol

    def lookup(self, identifier):
        if (identifier not in self.symbols):
            raise RuntimeException(f'{identifier} not defined')
        return self.symbols[identifier]

    def to_dict(self):
        return self.symbols


class Function(SymbolTable):

    def __init__(self, is_builtin=False):
        super().__init__()
        self.is_builtin = is_builtin


class Symbol:

    def __init__(self, identifier, node):
        self.identifier = identifier
        self.node = node

    def get_identifier(self):
        return self.identifier

    def get_node(self):
        return self.node

    def is_function(self):
        return isinstance(self.node, Function)



class SyntaxException(Exception):
    
    def __init__(self, token):
        super().__init__(f'{token.value}, line {token.lineno}')


class RuntimeException(Exception):
    
    def __init__(self, *args: object) -> None:
        super().__init__(*args)
